/*   
FlobSet.java
 *    
 *    Copyright (c) 2000, Ted Nelson and Tuomas Lukka
 *
 *    You may use and distribute under the terms of either the GNU Lesser
 *    General Public License, either version 2 of the license or,
 *    at your choice, any later version. Alternatively, you may use and
 *    distribute under the terms of the XPL.
 *
 *    See the LICENSE.lgpl and LICENSE.xpl files for the specific terms of 
 *    the licenses.
 *
 *    This software is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the README
 *    file for more details.
 *
 */
/*
 * Written by Tuomas Lukka
 */
package org.gzigzag;
import java.util.*;
import java.awt.*;

/** A set of flobs, i.e.things that are rendered to screen associated
 * with cells.
 * Flobs are stored with depths, which are stably sorted in the end 
 * so that the flobs placed later are on top.
 */

public class FlobSet implements ZZScene {
public static final String rcsid = "$Id: FlobSet.java,v 1.29 2000/12/11 02:54:44 tjl Exp $";

    public static final boolean dbg = true;
    static final void p(String s) { if(dbg) System.out.println(s); }
    static final void pa(String s) { System.out.println(s); }

    Dimension size;
    public Dimension getSize() {
	return size;
    }
    Color fg, bg;
    Component cpt;

    FlobSet(Dimension size, Color foreground, Color background, 
		Component cpt) {
	this.size = size;
	this.bg = background;
	this.fg = foreground;
	this.cpt = cpt;
    }
    public Component getComponent() { return cpt; }
    public Color getBackground() { return bg; }


    private Hashtable c2ph;

    private final Hashtable getPH(Flob parent) {
	Object key;
	if(parent == null) key = "";
	else key = parent;
	Hashtable res = (Hashtable)c2ph.get(key);
	if(res == null) 
	    c2ph.put(key, res=new Hashtable());
	return res;
    }

    private int nflobs;
    private Flob[] flobs = new Flob[30];
    boolean fsorted = false;
    boolean ftagged = false;

    
    private int nrend;
    /** Decorations.
     */
    private Renderable[] rends = new Renderable[10];
    boolean rsorted = false;

    private int maxdepth = 0;
    
    public void add(Renderable r) {
	// Check: if it happens to be a flob, then put
	// it as a flob.
	if(r instanceof Flob) {
	    add((Flob)r);
	    return;
	}
	
	if(r.d > maxdepth) maxdepth = r.d;
	if(nrend >= rends.length) {
	    Renderable[] n = new Renderable[nrend + 50];
	    System.arraycopy(rends, 0, n, 0, nrend);
	    rends = n;
	}
	rends[nrend++] = r;
	rsorted = false;
    }

    /** Add a flob into the flobset.
     * Note that flobPath might not be set yet.
     * It is therefore used only when the flobs are being sorted.
     */
    public void add(Flob f) {
	// p("AddFlob "+f.x+" "+f.y+" "+f.w+" "+f.h+" "+f.d);
	if(f.d <= 0) 
	    throw new ZZError("Negative or zero depth for flob!");
	if(f.d > maxdepth) maxdepth = f.d;
	// if(f.c != null) 
	//    getPH(f.flobPath).put(f.c, f);
	if(nflobs >= flobs.length) {
	    Flob[] n = new Flob[nflobs + 100];
	    System.arraycopy(flobs, 0, n, 0, nflobs);
	    flobs = n;
	}
	flobs[nflobs++] = f;
	fsorted = false;
	ftagged = false;
    }

    public final void tag() {
	if(ftagged) return;
	c2ph = new Hashtable();
	for(int i=0; i<nflobs; i++) 
	  if(flobs[i].c != null) {
	    Hashtable h = getPH(flobs[i].getParent());
	    Vector v = (Vector)h.get(flobs[i].c);
	    if(v==null) h.put(flobs[i].c, v = new Vector());
	    v.addElement(flobs[i]);
	  }
	ftagged = true;
    }

    public final void sort() { sortf(); sortr(); }
    public final void sortf() { 
	if(fsorted) return;
	// pa("SORTING");
	// dump();
	Renderable.depthSort(flobs, nflobs); 
	fsorted = true;
	// dump();
    }
    public final void sortr() { 
	if(rsorted) return;
	Renderable.depthSort(rends, nrend); 
	rsorted = true;
    }

    protected void depthColor(Graphics g, int depth) {
	// XXX Foreground adjustable?
	if(depth > maxdepth) 
	    throw new ZZError("Too strange. "+depth+" "+maxdepth);
	int fract = (256*(depth-1)) / (maxdepth + maxdepth / 10);
	Color ncol = ZZUtil.mix(fg, bg, fract);
	g.setColor(ncol);
	// p("Color: "+fract+" "+d+" "+maxdepth+" "+ncol);
    }

    public void render(Graphics g) {

	// ZZDrawing.instance.setAlpha(g, (float)0.5);
	ZZDrawing.instance.setQuality(g);

	int f=nflobs-1;
	int r=nrend-1;
	int oldd = -1;
	sort();
	// p("StartColors");
	while(f >= 0 || r >= 0) {
	    // Stagger the two arrays. Always render decorations of the same
	    // depth last
	    while(f >= 0 && 
	     ( r < 0 || rends[r].d <= flobs[f].d)) {
	        if(flobs[f].d != oldd)
		    depthColor(g, oldd = flobs[f].d);
		flobs[f--].render(g);
	    }
	    while(r >= 0 && 
	     ( f < 0 || rends[r].d > flobs[f].d)) {
	        if(rends[r].d != oldd)
		    depthColor(g, oldd = rends[r].d);
		rends[r--].render(g);
	    }
	}
	renderDC(g);
    }

    FlobSet interpPrepared = null;
    public void renderInterp(Graphics g, ZZScene other0, float fract) {
	sort();
	FlobSet other = (FlobSet)other0;
	if(interpPrepared != other) {
	    prepareInterp(other);
	}
	renderInterp(g, fract);
    }

    public void renderInterp(Graphics g, float fract) {
	int oldd = -1;
	for(int i=nflobs-1; i>=0; i--) {
	    if(flobs[i].d != oldd)
		depthColor(g, oldd = flobs[i].d);
	    flobs[i].renderInterp(g, fract);
	}
    }

    public boolean isInterpUseful(ZZScene other0) {
	FlobSet other = (FlobSet)other0;
	return prepareInterp(other);
    }

    /** This is not nice - vectors are too heavyweight.
     */
    public Vector findFlobs(Flob parent, ZZCell c) {
	tag();
	return (Vector)getPH(parent).get(c);
    }

    /** True if from's center is nearer to test's than other's. */
    static boolean isNearer(Flob from, Flob test, Flob other) {
	int fx = from.x + from.w/2, fy = from.y + from.h/2;
	int tx = test.x + test.w/2, ty = test.y + test.h/2;
	int ox = other.x + other.w/2, oy = other.y + other.h/2;
	return ((fx-tx)*(fx-tx) + (fy-ty)*(fy-ty))
	     < ((fx-ox)*(fx-ox) + (fy-oy)*(fy-oy));
    }
	

    /** Find the flob in this flobset which is nearest to a flob in another
     * flobset -- i.e., f is supposed NOT to be in this flobset. Distances
     * are measured center to center.
     * @param f The flob to search from -- should NOT be in this set.
     * @param parent The parent of the flob to find -- must be in THIS set.
     */
    public Flob findNearest(Flob f, Flob parent) {
	Vector flobs = findFlobs(parent, f.c);
	if(flobs==null) return null;
	Flob result = null;
	for(Enumeration e = flobs.elements(); e.hasMoreElements();) {
	    Flob g = (Flob)e.nextElement();
	    if(result == null || isNearer(f, g, result))
		result = g;
	}
	if(result != null) return result;
	else return null;
    }

    /** For compatibility
     * deprecated Use findFlobs, findNearest or something.
     */
    public Flob findFlob(String path, ZZCell c) {
	Vector flobs = findFlobs(null, c);
	if(flobs == null) return null;
	return (Flob)flobs.elementAt(0);
    }

    public boolean prepareInterp(FlobSet f2) {
	boolean useful = false;
	for(int i=0; i<f2.nflobs; i++)
	    f2.flobs[i].interpTo = null;
for(int iter=0; iter<5; iter++) {
	for(int i=0; i<nflobs; i++) {
	    if(flobs[i].c == null) continue;
	    Flob parent = flobs[i].getParent();
	    Flob f = f2.findNearest(flobs[i], 
			            parent != null ? parent.interpTo : null);
	    if(f==null ||
	       (f.interpTo != null && isNearer(f, f.interpTo, flobs[i]))) {
		// First set to null -- avoid animation errors
		flobs[i].interpTo = null;
		continue;
	    }
	    flobs[i].interpTo = f;
	    f.interpTo = flobs[i];
	    useful = (useful || flobs[i].isInterpUseful());
	}
}
	f2.interpPrepared = this;
	interpPrepared = f2;
	return useful;
    }

    public Object getObjectAt(int x, int y) { 
	return getObjectAt(x, y, null);
    }

    public Object getObjectAt(int x, int y, Flob[] fres) { 
	sortf();
	for(int i=0; i<nflobs; i++) {
	    Object o = flobs[i].hit(x, y);
	    if(o != null) {
		if(fres != null) fres[0] = flobs[i];
		return o;
	    }
	}
	return null;
    }

    public void renderXOR(Graphics g, int x, int y) {
	Flob[] which = new Flob[1];
	Object co = getObjectAt(x, y, which);
	if(co == null) return;
	which[0].renderXOR(g, x, y);
    }

    public void dump() {
	pa("FlobSet dump: "+nflobs+" "+nrend);
	for(int i=0; i<flobs.length; i++) {
	    if(flobs[i] == null) {
		pa(" Flob null!");
		continue;
	    }
	    pa(" Flob "+flobs[i].x+" "+flobs[i].y+" "+
		flobs[i].w+" "+flobs[i].h+" "+flobs[i].d+" "+flobs[i]);
	}
	for(int i=0; i<rends.length; i++) {
	    if(rends[i] == null) {
		pa(" Renderable null!");
		continue;
	    }
	    pa(" Renderable "+ rends[i].d+" "+rends[i]);
	}
    }

    /** A callback interface for iterating over the Flobs
     * in a FlobSet in depth order.
     */
    public interface DepthIter {
	void act(Flob[] flobs, int start, int n);
    }
    /** Iterate a routine over a blocks of same-depth flobs.
     */
    public void iterDepth(DepthIter di, boolean frontFirst) {
	if(nflobs <= 0) return;
	if(flobs == null) throw new Error("NULL FLOBS");
	if(flobs[0] == null) throw new Error("NULL FIRST FLOB");
	sortf();
	if(flobs == null) throw new Error("NULL FLOBS AFTER SORT");
	if(flobs[0] == null) throw new Error("NULL FIRST FLOB AFTER SORT");
	if(frontFirst) {
	    int cur = 0; int curd = flobs[0].d;
	    for(int i=0; i<nflobs; i++) {
		if(curd != flobs[i].d ) {
		    di.act(flobs, cur, i-cur);
		    cur = i;
		    curd = flobs[i].d;
		}
	    }
	    di.act(flobs, cur, nflobs-cur);
	} else {
	    int cur = nflobs-1; int curd = flobs[nflobs-1].d;
	    for(int i=nflobs-1; i>=0; i--) {
		if(curd != flobs[i].d) {
		    di.act(flobs, i+1, cur-i);
		    cur = i;
		    curd = flobs[i].d;
		}
	    }
	    di.act(flobs, 0, cur+1);
	}
    }

    DragCursor dc; int dcx, dcy;
    public void setDragCursor(DragCursor dc, int x, int y) {
	this.dc = dc; this.dcx = x; this.dcy = y;
    }
    public interface DragCursor {
	boolean accept(Object o);
    }

    /** Render the current dragCursor.
     */
    private void renderDC(Graphics g) {
	if(dc == null) return;
	Flob[] which = new Flob[1];
	Object co = getObjectAt(dcx, dcy, which);
	if(co == null) return;
	if(!dc.accept(co)) return;
	which[0].renderXOR(g, dcx, dcy);
    }

}
